# Carousel version of call_function

#define FROM_ASM
#include "rtos.h"

#define MSR_CACHES 0x80


.global call_function
.global init_asm_start
.global kernel_entry
.global return_point
.global spm_loader
.global inspection

	.section	.kernelstuff.d,"ax",@progbits
shadow_area:
        .space (NUM_BLOCKS * 4)

	.section	.kernelstuff.t,"ax",@progbits

dma_spm = DMA_SPM_START

call_function:
        # r5: address of target function
        # r6: block info word
        # r7-r10: parameters of target function
        # r15: return PC - 8

        # SYSCALL mode: disable interrupts, breaks
        msrclr  r12, MSR_IE | MSR_BE
    
        .long   CMD_CALL

        # Set spm_index for S^spm - do Save
        put     r6, rfsl0
        addi    r18, r0, 1
        cput    r18, rfsl0

        # Set spm_index for S_r0 (first region size)
        bsrli   r11, r6, 16
        put     r11, rfsl0

        # Start location for load
        add     r11, r5, r11

0:      # Do Load
        cput    r11, rfsl0
        addi    r11, r11, -BLOCK_SIZE
        rsub    r18, r5, r11
        bgei    r18, 0b

        # Wait for completion of final Load
        get     r18, rfsl0

# {\Execute}. Task $\tau_j$ is executed from SPM. Execution may 
# involve one or more region transitions; new regions are loaded into SPM
# as necessary. The maximum region size is $S^\spm_j$ blocks.

        swi     r15, r17, 0             # Push call_function return address 
        shi     r6, r17, 4              # Push r6 (BIW) low hword only
        swi     r12, r17, 8             # Push entry MSR state
        addi    r17, r17, 16
#ifdef CHECK
        swi     r5, r17, -4

        addi    r11, r6, 16-4           
        andni   r11, r11, 1
0:      lw      r3, r11, r5
        lwi     r4, r11, 0x10000000
        xor     r12, r4, r3
        bnei    r12, bad1
        addi    r11, r11, -4
        bnei    r11, 0b
#endif

        # ** Handover to called function **
        # USER mode: enable interrupts, breaks
        msrset  r3, MSR_IE | MSR_BE

        # Parameter registers moved into place
        or      r4, r5, r0
        or      r5, r7, r0
        or      r6, r8, r0
        or      r7, r9, r0

        # Call function
        bralid  r15, dma_spm
        #brald   r15, r4
        or      r8, r10, r0

return_point:

# {\Restore}. Upon completion, the original contents of the $S^\spm_j$
# SPM blocks are restored. 

        # SYSCALL mode: disable interrupts, breaks
        msrclr  r5, MSR_IE | MSR_BE

        .long   CMD_RTFF

        addi    r17, r17, -16

        lwi     r15, r17, 0         # Return point restored
        lhui    r6, r17, 4          # BIW restored -- low hword
        lwi     r12, r17, 8         # entry MSR state

        # Start Restore
        # Set spm_index
        put     r0, rfsl0

0:      # Do Restore
        cput    r0, rfsl0
        addi    r6, r6, -BLOCK_SIZE
        bgei    r6, 0b

        # Wait for completion of final Restore
        get     r5, rfsl0

#ifdef CHECK
        lwi     r11, r17, 4-16  # previous BIW
        lwi     r5, r17, 12-16 # previous function
        addi    r11, r11, 16-4
        andni   r11, r11, 1
0:      lw      r9, r11, r5
        lwi     r8, r11, 0x10000000
        xor     r9, r8, r9
        bnei    r9, bad2
        addi    r11, r11, -4
        bnei    r11, 0b
#endif

        rtsd    r15, 8
        # Entry status returned (enable interrupts if
        # returning to user code)
        mts     rmsr, r12

.text
rcv_source:
        # {\Load}. A new region of specified size is loaded into SPM.
        # r3: next region start in external memory
        # r5: size of region (BIW)
        braid   rcv_main
        add     r11, r3, r5

.section	.kernelstuff.t,"ax",@progbits
rcv_main:
        # Set spm_index
        put     r5, rfsl0

0:      # Do Load
        cput    r11, rfsl0
        addi    r11, r11, -BLOCK_SIZE
        rsub    r18, r3, r11
        bgei    r18, 0b

        # Wait for completion of final Load then start it
        get     r18, rfsl0
        brai    dma_spm     # Do not use delay slot (target not ready)

.text
spm_loader:
        # r5: SPM block number
        # r6: address in external memory
        # r7: BIW

        # Find endpoint
        add     r11, r6, r7

        # Set spm_index
        add     r7, r5, r7
        put     r7, rfsl0

0:      # Do Load
        cput    r11, rfsl0
        addi    r11, r11, -BLOCK_SIZE
        rsub    r18, r6, r11
        bgei    r18, 0b

        # Wait for completion of final Load, return
        rtsd    r15, 8
        get     r18, rfsl0

kernel_entry:
        # Init region change vector (3 words)
        lwi     r3, r0, rcv_source
        swi     r3, r0, REGION_CHANGE
        lwi     r3, r0, rcv_source + 4
        swi     r3, r0, REGION_CHANGE + 4
        lwi     r3, r0, rcv_source + 8
        swi     r3, r0, REGION_CHANGE + 8
    
        # Init shadow area pointer; start kernel
        addik   r17, r0, shadow_area
        braid   init_asm_start
        # SYSTEM mode: interrupts off, breaks on
        addi    r3, r0, MSR_BE

#if 0
bad1:
        addi    r7, r0, msg1
        bri     badc


bad2:
        addi    r7, r0, msg2
        addi    r17, r17, 16
        add     r3, r9, r0
        add     r4, r8, r0

badc:
        swi     r3, r1, -100
        swi     r4, r1, -104
        swi     r5, r1, -108
        swi     r6, r1, -112
        swi     r11, r1, -116
        bralid  r15, print
        or      r5, r7, r0

        bralid  r15, printhexs
        lwi     r5, r1, -100

        bralid  r15, printhexs
        lwi     r5, r1, -104

        bralid  r15, printhexs
        lwi     r5, r1, -108

        bralid  r15, printhexs
        lwi     r5, r1, -112

        bralid  r15, printhexs
        lwi     r5, r1, -116


        bralid  r15, printhexs
        lwi     r5, r17, -16 - 16

        bralid  r15, printhexs
        lwi     r5, r17, -12 - 16

        bralid  r15, printhexs
        lwi     r5, r17, -8 - 16

        bralid  r15, printhexs
        lwi     r5, r17, -4 - 16

        bralid  r15, printhexs
        lwi     r5, r17, -12

        addi    r5, r0, gap
        bralid  r15, print
        nop

        addi    r29, r0, 0
0:      lwi     r5, r29, 0x10000000
        bralid  r15, printhexs
        addi    r29, r29, 4

        xori    r5, r29, 0x10000020
        bnei    r5, 0b


        braid   inspection
        nop

        bralid  r15, exit
        nop

printhexs:
        or      r31, r0, r15
        or      r30, r0, r5
        addi    r5, r0, sp
        bralid  r15, print
        nop
        bralid  r15, printhex
        or      r5, r30, r0
        rtsd    r31, 8
        nop

.data
sp:
        .ascii " "
        .byte 0

msg1:
        .ascii "cond1\n"
        .byte 0
msg2:
        .ascii "cond2\n"
        .byte 0
gap:
        .ascii "\n"
        .byte 0

#endif
