#include "rtos.h"

#define SPM_TESTS           16
#define CACHE_TESTS         32

extern void inspect_sp(void);

static unsigned rng(void)
{
    static unsigned m_w = 0xdead;    /* must not be zero */
    static unsigned m_z = 0x31337;    /* must not be zero */
    m_z = 36969 * (m_z & 65535) + (m_z >> 16);
    m_w = 18000 * (m_w & 65535) + (m_w >> 16);
    return (m_z << 16) + m_w;  /* 32-bit result */
}

typedef unsigned short (* FN1) (unsigned short, unsigned char);

static unsigned short icrc(unsigned short crc, unsigned char onech)
{
	int             i;
	unsigned short  ans = (crc ^ onech << 8);

	for (i = 0; i < 8; i++) {
		if (ans & 0x8000) {
            ans = (ans << 1);
			ans = ans ^ 4129;
		} else {
			ans <<= 1;
        }
	}
	return ans;
}

#define NSORT 50
unsigned search_size;

unsigned sort_array[NSORT] AKD;
static unsigned base_table[NSORT];

void insertsort(void)
{
	unsigned i, j, temp;

	i = 1;
	while (i < NSORT) {
		j = i;
		while ((j > 0) && (sort_array[j] < sort_array[j - 1])) {
			temp = sort_array[j];
			sort_array[j] = sort_array[j - 1];
			sort_array[j - 1] = temp;
			j--;
		}
		i++;
	}
}

extern void jack(void);

void cjack (void)
{
    unsigned i, j;
/*
    for (i = 0; i < NSORT; i++) {
        //sort_array[i] *= i + 1;
    }
*/
    j = 0;
    for (i = 0; i < NSORT; i++) {
       // sort_array[i] = 
        j = (sort_array[i] + j);
        sort_array[i] = j;
    }
/*
    j = 0;
    for (i = 1; i < NSORT; i++) {
        if (sort_array[i] > sort_array[i-1]) {
            j++;
        }
    }
    sort_array[0] = j;
    for (i = 0; i < NSORT; i++) {
        //sort_array[i] /= i + 1;
    }
*/
}

static void do_measure(FN1 crc_src)
{
    unsigned i, j, a, b, c;
    FN1 crc_dest = (FN1) DMA_SPM_START;

    spm_loader(0, (FN)crc_src, DEFAULT_LOAD_BIW);
    for (i = 0; i < 4; i++) {
        for (j = 0; j < NSORT; j++) {
            sort_array[j] = base_table[j];
        }
        a = get_time();
        b = get_time();
        j = crc_dest(j, i);
        c = get_time();
    }
    print("spm: ");
    printhex((c - b) - (b - a));

    for (i = 0; i < 4; i++) {
        for (j = 0; j < NSORT; j++) {
            sort_array[j] = base_table[j];
        }
        a = get_time();
        b = get_time();
        j = crc_src(j, i);
        c = get_time();
    }
    print("  cache: ");
    printhex((c - b) - (b - a));
    print("\n");
}

extern unsigned test_chain[];

extern unsigned region_change_timer(void);

extern unsigned null_call(unsigned spm_block, FN ext_mem_src, unsigned size);

extern unsigned tgt_function(void);


static void measure_for(unsigned sspm, unsigned sr0,
                unsigned * enter, unsigned * leave)
{
    unsigned q = BIW(sspm, sr0);
    unsigned a, b, a0, i;
    unsigned e, l, c = 0;
    unsigned et, lt;
    
    et = lt = 0;

    for (i = 0; i < 110; i++) {
        a = get_time();
        a0 = call_function(tgt_function, q, 0, 0, 0, 0);
        b = get_time();

        if (i >= 10) {
            e = a0 - a;
            l = b - a0;
            et += e;
            lt += l;
            c++;
        }
    }
    (* enter) = (et + (c / 2)) / c;
    (* leave) = (lt + (c / 2)) / c;
}

static unsigned call_2_overhead, return_2_overhead, brt_spm;

static void cf_timer(void)
{
    unsigned cf2 = DEFAULT_CALL_BIW;
    unsigned e0, l0, e1, l1, e2, l2;
    unsigned e_sspm_scale, l_sspm_scale, e_sr0_scale, l_sr0_scale;
    unsigned i, j, e_base, l_base;
    unsigned e_pessim = 0;
    unsigned l_pessim = 0;
    unsigned cache = 0;
    unsigned * ext_mem_loc;
    const unsigned SC = 5;
    const unsigned XD = 5;


    /* Initial step - get estimates */

    while(1) {
        print("# estimate\n");

        /* base case */
        call_function(measure_for, cf2, SC, SC, &e0, &l0);

        /* consequence of adding 1 to S^spm */
        call_function(measure_for, cf2, XD+SC, SC, &e1, &l1);
        e_sspm_scale = e1 - e0;
        l_sspm_scale = l1 - l0;

        /* consequence of adding 1 to S_r_0 */
        call_function(measure_for, cf2, XD+SC, XD+SC, &e2, &l2);
        e_sr0_scale = e2 - e1;
        l_sr0_scale = l2 - l1;

        if (l_sr0_scale != 0) {
            print("# l_sr0_scale should be 0, but is ");
            printdec(l_sr0_scale);
            print("\n");
            continue;
        }


        e_sspm_scale /= XD;
        l_sspm_scale /= XD;
        e_sr0_scale /= XD;

        if (e_sr0_scale != l_sspm_scale) {
            print("# BRT mismatch: ");
            printdec(e_sr0_scale);
            print(" != ");
            printdec(l_sspm_scale);
            print("\n");
            if (e_sr0_scale < l_sspm_scale) {
                e_sr0_scale = l_sspm_scale;
            }
            l_sspm_scale = e_sr0_scale;
            continue;
        }
        if (e_sspm_scale == 0) {
            print("# C^Save_1 operation is not really free\n");
            continue;
        }

        e_base = e0 - (e_sspm_scale * SC) - (e_sr0_scale * SC);
        l_base = l0 - (l_sspm_scale * SC);

        if (((int) e_base) <= 0) {
            print("# e_base too small? -");
            printdec(-e_base);
            print("\n");
            continue;
        }
        if (((int) l_base) <= 0) {
            print("# l_base too small? -");
            printdec(-l_base);
            print("\n");
            continue;
        }
        break;
    }

    /* further testing; determine pessimism in the estimates */
    print("# pessimism?\n");

    for (j = 0; j < search_size; j++) {
        unsigned q = rng();
        unsigned sspm = (q & 0x3f) + 1;
        unsigned sr0 = ((q >> 6) % sspm) + 1;
        call_function(measure_for, cf2, sspm, sr0, &e2, &l2);

        e1 = e_base + (e_sspm_scale * sspm) + (e_sr0_scale * sr0);
        l1 = l_base + (l_sspm_scale * sspm);

        if ((e2 > e1) && ((e2 - e1) > e_pessim)) {
            e_pessim = e2 - e1;
        }
        if ((l2 > l1) && ((l2 - l1) > l_pessim)) {
            l_pessim = l2 - l1;
        }
    }

    /* What's the cache miss time? */
    print("# cache miss?\n");

    for (j = 0; j < search_size; j++) {
        unsigned a, b, c;

        asm volatile(
                "addi   r18, r0, tgt_function\n"
                "msrclr r0, 0x20\n" /* IC off */
                "wic    r18, r18\n"
                "msrset r0, 0x20\n" /* IC on */

                "nop\nnop\nnop\nnop\n"

                "brald r15, r18\n"  // First call, cache miss
                "get    %0, rfsl1\n"  

                "brald r15, r18\n"  // Second call, cache hit
                "get    %1, rfsl1\n"

                "get    %2, rfsl1\n" : "=r"(a), "=r"(b), "=r"(c) );

        c = (b - a) - (c - b);
        cache += c;
    }
    cache /= search_size;

    /* output */
    print("# Constants determined through ");
    printdec(search_size);
    print(" tests:\n");
    print("SAVE_PER_BLOCK = ");
    printdec(e_sspm_scale * CLOCK_PERIOD);
    print("#ns\nBRT_SPM = ");
    printdec(e_sr0_scale * CLOCK_PERIOD);
    print("#ns\nSAVE_OVERHEAD = ");
    printdec((e_base + e_pessim) * CLOCK_PERIOD);
    print("#ns\nRESTORE_OVERHEAD = ");
    printdec((l_base + l_pessim) * CLOCK_PERIOD);
    print("#ns\nBRT_CACHE = ");
    printdec(cache * CLOCK_PERIOD);
    print("#ns\n");

    brt_spm = e_sr0_scale;
    call_2_overhead = e_base + (e_sspm_scale * 2) + (e_sr0_scale * 2);
    return_2_overhead = l_base + (l_sspm_scale * 2);
}


static unsigned c_region_change_timer(unsigned region_size)
{
    unsigned x;

    x = call_function(region_change_timer, 
        DEFAULT_CALL_BIW,
        (region_size - 1) * BLOCK_SIZE, 0, 0, 0);
    return x;
}

#define RC_BLOCKS 5

static void measure_region_change(void)
{
    unsigned minimum = 1 << 31;
    unsigned maximum = 0;
    unsigned j;

    for (j = 0; j < search_size; j++) {
        unsigned x = (rng() & 0xf) + 1; // number of blocks
        unsigned rx = c_region_change_timer(x);

        rx -= brt_spm * x;
        
        if (rx < minimum) {
            minimum = rx;
        }
        if (rx > maximum) {
            maximum = rx;
        }
    }

    print("# C^load_1 = BRT^spm = ");
    printdec(brt_spm * CLOCK_PERIOD);
    print("ns\nTRANSITION_WCET_OVERHEAD = ");
    printdec(maximum * CLOCK_PERIOD);
    print("#ns\n");
}

int main(void)
{
    unsigned i;

    print("======================\n");
    print("measure.c running\n");
    print("======================\n");


    asm volatile("msrset r18, 0x20"); /* IC on */
    asm volatile("msrclr r18, 0x80"); /* DC off */

    search_size = 10;
    for (i = 0; i < 5; i++) {
        set_trigger(1<<31);
        print("\n\n######################\n");
        call_function(cf_timer, 
            DEFAULT_CALL_BIW, 0, 0, 0, 0);

        print("# measure_region_change\n");
        measure_region_change();

        print("# cs_measure\n");
        cs_measure(15, search_size, call_2_overhead, return_2_overhead);

        search_size *= 10;
    }
    print("\n\n######################\n");
    print("#completed measurement\n");

    return 0;
}

